#ifndef ERDPROPERTIESPANEL_H
#define ERDPROPERTIESPANEL_H

class ErdPropertiesPanel
{
    public:
        virtual bool commitErdChange() = 0;
        virtual void abortErdChange() = 0;
};

#endif // ERDPROPERTIESPANEL_H
