table
{
    border-style: solid;
    border-width: 1px;
    border-color: black;
    border-collapse: collapse;
}

table tr
{
    background-color: white;
}

table tr.header
{
    background-color: #DDDDDD;
}

table tr.title
{
    background-color: #EEEEEE;
}

table tr td
{
    padding: 0px 3px 0px 3px;
    border-style: solid;
    border-width: 1px;
    border-color: #666666;
}

table tr td.null
{
    color: #999999;
    text-align: center;
    padding: 0px 3px 0px 3px;
    border-style: solid;
    border-width: 1px;
    border-color: #666666;
}

table tr td.separator
{
    padding: 0px 3px 0px 3px;
    border-style: solid;
    border-width: 1px;
    border-color: #666666;
    background-color: #DDDDDD;
}

table tr td.rownum
{
    padding: 0px 3px 0px 3px;
    border-style: solid;
    border-width: 1px;
    border-color: #666666;
    background-color: #DDDDDD;
    text-align: right;
}

th {
    cursor: pointer;
    text-decoration: underline;
    border-style: solid;
    border-width: 1px;
    border-color: black;
}

tr.data:nth-child(even) {
    background-color: Beige;
}

th.sorted-asc::after {
    content: " ▲";
}

th.sorted-desc::after {
    content: " ▼";
}
