#ifndef BOOLTOOLBUTTON_H
#define BOOLTOOLBUTTON_H

#include <QToolButton>


class BoolToolButton : public QToolButton
{
        Q_OBJECT

    public:
        BoolToolButton(QWidget *parent = nullptr);
};

#endif // BOOLTOOLBUTTON_H
